# -*- coding: utf-8 -*-

# Copyright (C) 2011 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Breaking lines to a specified width."""

import aeidon
import re
import sys

__all__ = ("Liner",)


class Liner(aeidon.Parser):

    """
    Breaking lines to a specified width.

    :ivar length_func: A function that returns the length of its argument
    :ivar max_length: Maximum length of a line in units of :attr:`length_func`
    :ivar max_lines: Maximum preferred amount of lines (may be exceeded)
    :ivar _penalties: List of penalty pattern dictionaries
    """

    # Reading Donald E. Knuth and Michael F. Plass's "Breaking Paragraphs into
    # Lines" from "Software--Practice and Experience" vol. 11 from 1981 is
    # recommended to understand the general problem of breaking a paragraph
    # of text into lines, the terminology used here as well of boxes, penalties
    # and demerits and the computational complexity of finding the optimal
    # solution in each case. Subtitling is in many ways a different
    # application, which is a lot simpler, but also more ambiguous and
    # subjective in terms of the minimized demerit measure.

    _re_multi_space = re.compile(r" {2,}")

    def __init__(self, re_tag=None, clean_func=None):
        """Initialize a :class:`Liner` instance."""
        aeidon.Parser.__init__(self, re_tag, clean_func)
        self._penalties = []
        self.length_func = len
        self.max_length = 40
        self.max_lines = 3

    def _boxes_to_lines(self, boxes, breaks):
        """Return `boxes` joined to form lines."""
        edges = [0] + [x + 1 for x in breaks] + [len(boxes)]
        return  [" ".join(boxes[edges[i]:edges[i+1]])
                 for i in range(len(edges) - 1)]

    def _break_lines(self, boxes, penalties, nlines):
        """
        Break `boxes` into lines and return break points and demerit.

        If keeping all boxes on a single line results in a valid and better
        result than splitting to `nlines` return an empty list. If no
        valid break points can be found, return ``None``.
        """
        breaks = self._list_possible_breaks(boxes, penalties, nlines)
        best_breaks = None
        best_demerit = sys.maxsize
        text = " ".join(boxes)
        if self.length_func(text) <= self.max_length:
            # Use a valid one-line solution as the benchmark,
            # that any more-line solution must beat.
            best_breaks = []
            best_demerit = self._calculate_demerit(boxes, penalties, [])
        if nlines == 1:
            return best_breaks, best_demerit
        if nlines == 2:
            for i in breaks:
                if penalties[i] > best_demerit: break
                demerit = self._calculate_demerit(boxes, penalties, [i])
                if demerit < best_demerit:
                    best_breaks = [i]
                    best_demerit = demerit
            return best_breaks, best_demerit
        # For more than two lines, loop over first break points and
        # recursively figure out rest of the breaks in each case.
        for i in breaks:
            # Use the maximum total negative penalty
            # that can accumulate from all later breaks.
            negpen = sorted(x for x in penalties[i+1:] if x < 0)
            negpen = sum(negpen[:min(len(negpen), nlines - 2)])
            if (penalties[i] + negpen) > best_demerit: break
            value = self._break_lines(boxes[i+1:], penalties[i+1:], nlines-1)
            if value[0] is None: continue
            later = [i + 1 + x for x in value[0]]
            demerit = self._calculate_demerit(boxes, penalties, [i] + later)
            if demerit < best_demerit:
                best_breaks = [i] + later
                best_demerit = demerit
        return best_breaks, best_demerit

    def break_lines(self):
        """Break lines and return text."""
        self.text = self.text.replace("\n", " ")
        self.pattern = self._re_multi_space
        self.replacement = " "
        self.replace_all()
        boxes = self.text.split(" ")
        if len(boxes) == 1:
            return self.get_text()
        penalties = self._detect_penalties(boxes)
        best_breaks = None
        best_demerit = sys.maxsize
        # We can probably handle up to ten lines of text
        # before finding break points gets intolerably slow.
        min_nlines = min(2, self.max_lines)
        max_nlines = min(10, len(boxes))
        for nlines in range(min_nlines, max_nlines + 1):
            breaks, demerit = self._break_lines(boxes, penalties, nlines)
            if breaks is None: continue
            if demerit < best_demerit:
                best_breaks = breaks
                best_demerit = demerit
            if nlines < self.max_lines:
                continue
            pos = -1
            for i in range(len(boxes)):
                pos = pos + 1 + len(boxes[i])
                text = self.text
                if i in best_breaks:
                    text = text[:pos] + "\n" + text[pos+1:]
                self.text = text
            return self.get_text()
        # If text cannot be broken, return original text.
        return self.get_text()

    def _calculate_demerit(self, boxes, penalties, breaks):
        """Return demerit measure for `boxes` broken by `breaks`."""
        nlines = len(breaks) + 1
        penalties = [penalties[i] for i in breaks]
        lines = self._boxes_to_lines(boxes, breaks)
        lengths = list(map(self.length_func, lines))
        mlength = sum(lengths) / len(lengths)
        xlength = self.max_length
        # Use two subjective measures of badness: (1) 'deviation',
        # which is the variance of line lengths relative to the
        # maximum line length and (2) upside-down 'pyramid', which
        # is the sum of how much longer each line is than the next.
        return (sum(penalties)
                + 50 * sum(((x - mlength) / xlength)**2 for x in lengths)
                + 50 * sum(((lengths[i] - lengths[i+1]) / xlength)**2
                           for i in range(len(lengths) - 1)
                           if lengths[i] > lengths[i+1])

                + 100 * (nlines-1)**3
                + 1000 * max(0, nlines - self.max_lines)**3)

    def _detect_penalties(self, boxes):
        """Detect penalties for break points following `boxes`."""
        text = " ".join(self._boxes_to_lines(boxes, breaks=[]))
        textpen = [0] * len(text)
        for penalty in self._penalties:
            self.pattern = penalty["regex"]
            self.pos = 0
            while True:
                try:
                    self.next()
                except StopIteration:
                    break
                start, end = self.match.span(penalty["group"])
                # Use sum, since in some rare cases multiple
                # patterns can match the same space.
                textpen[start] += penalty["value"]
        penalties = [0] * len(boxes)
        pos = -1
        for i in range(len(boxes) - 1):
            pos = pos + 1 + len(boxes[i])
            penalties[i] = textpen[pos]
        return penalties

    @aeidon.deco.memoize(100)
    def _list_possible_breaks(self, boxes, penalties, nlines):
        """
        Return a list of all possible break points for `boxes`.

        All break points that would necessarily cause `max_length` to be
        violated are discarded. Breaks are returned sorted in ascending order
        of associated `penalties`, so that all remaining breaks can be
        discarded once a demerit threshold is crossed.
        """
        breaks = list(range(len(boxes) - (nlines - 1)))
        breakpen = penalties[:len(breaks)]
        if nlines == 1:
            return []
        if nlines == 2:
            keep = [False] * len(breaks)
            for i in range(len(breaks)):
                lines = self._boxes_to_lines(boxes, breaks=[i])
                lengths = map(self.length_func, lines)
                keep[i] = max(lengths) <= self.max_length
            breaks = [breaks[i] for i in range(len(breaks)) if keep[i]]
            breakpen = [breakpen[i] for i in range(len(breakpen)) if keep[i]]
            # Sort breaks in ascending order by penalties,
            # so that all remaining breaks can be discarded once
            # a demerit threshold is crossed.
            if not breaks: return []
            points = sorted(zip(breakpen, breaks))
            breakpen, breaks = zip(*points)
            return(breaks)
        # For more than two lines, loop over first break points and
        # recursively figure out rest of the breaks in each case.
        keep = [False] * len(breaks)
        for i in range(len(breaks)):
            lines = self._boxes_to_lines(boxes, breaks=[i])
            alength = self.length_func(lines[0])
            if alength > self.max_length: break
            later = self._list_possible_breaks(boxes[i+1:],
                                               penalties[i+1:],
                                               nlines-1)

            keep[i] = bool(later)
        breaks = [breaks[i] for i in range(len(breaks)) if keep[i]]
        breakpen = [breakpen[i] for i in range(len(breakpen)) if keep[i]]
        # Sort breaks in ascending order by penalties,
        # so that all remaining breaks can be discarded once
        # a demerit threshold is crossed.
        if not breaks: return []
        points = sorted(zip(breakpen, breaks))
        breakpen, breaks = zip(*points)
        return(breaks)

    def set_penalties(self, penalties):
        """
        Set penalty patterns.

        `penalties` should be a list of dictionaries with items "pattern",
        "flags", "group" and "value", where pattern is a regular expression
        with group parentheses around a space, group is the number of the group
        in pattern to hold the penalty of value. A negative penalty encourages
        a break and a positive penalty discourages.
        """
        self._penalties = []
        for penalty in penalties:
            regex = re.compile(penalty["pattern"], penalty["flags"])
            self._penalties.append(dict(regex=regex,
                                        group=penalty["group"],
                                        value=penalty["value"]))

    def set_text(self, text):
        """Set the target text to search in and parse it."""
        aeidon.Parser.set_text(self, text.strip())
        self.text = self.text.strip()



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Text markup for the Advanced Sub Station Alpha format."""

import aeidon

__all__ = ("AdvSubStationAlpha",)


class AdvSubStationAlpha(aeidon.markups.SubStationAlpha):

    """
    Text markup for the Advanced Sub Station Alpha format.

    In addition to the markup in Sub Station Alpha, Advanced Sub Station Alpha
    contains a whole lof of markup tags of which the following are of interest
    to us. The further complicated color tags that define numbered (?) colors
    and alpha channels are ignored. The reset tag is allowed a style definiton,
    e.g. ``{\\rDefault}`` to revert to style "Default".

     * ``{\\bWEIGHT}...{\\b0}``
     * ``{\\u1}........{\\u0}``
     * ``........{\\r[STYLE]}``
    """

    format = aeidon.formats.ASS

    _closing_pattern = r"\{\\([biu])0\}"
    _opening_pattern = r"\{\\(?![biu]0)(b|i|u|c|fn|fs).*?\}"
    _reset_pattern   = r"\{\\r.*?\}"

    def _main_decode(self, text):
        """Return `text` with decodable markup decoded."""
        text = self._decode_b(text, r"\{\\b[1-9]\d*\}(.*?)\{\\b[0\\]\}", 1)
        text = self._decode_u(text, r"\{\\u1\}(.*?)\{\\u[0\\]\}", 1)
        return aeidon.markups.SubStationAlpha._main_decode(self, text)

    def underline(self, text, bounds=None):
        """Return underlined `text`."""
        a, z = bounds or (0, len(text))
        target = "{{\\u1}}{}{{\\u0}}".format(text[a:z])
        return "".join((text[:a], target, text[z:]))



# -*- coding: utf-8 -*-

# Copyright (C) 2009 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Cell renderer for float data with fixed precision."""

import aeidon
import gaupol
import re

from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gtk

__all__ = ("FloatCellRenderer",)


class FloatCellRenderer(Gtk.CellRendererText):

    """Cell renderer for float data with fixed precision."""

    __gtype_name__ = "FloatCellRenderer"

    def __init__(self, format="{:.3f}"):
        """Initialize a :class:`FloatCellRenderer` instance."""
        GObject.GObject.__init__(self)
        self._format = format
        aeidon.util.connect(self, self, "notify::text")
        aeidon.util.connect(self, self, "edited")
        aeidon.util.connect(self, self, "editing-started")

    def _on_edited(self, renderer, path, text):
        """Check `text` for validity before sending onwards."""
        if re.match(r"^[+-]?(\d+[,.]?\d*|\d*[,.]?\d+)$", text) is None:
            renderer.stop_emission("edited")
            renderer.stop_editing(True)

    def _on_editing_started(self, renderer, editor, path):
        """Set `editor` to use same font as `self`."""
        gaupol.style.use_font(editor, "custom")

    def _on_notify_text(self, *args):
        """Cut decimals to fixed precision."""
        # Since GTK 3.6, the notify::text signal seems to get
        # emitted insanely often even if text hasn't changed at
        # all. Let's try to keep this callback as fast as possible.
        self.props.markup = self._text_to_markup(self.props.text)

    @aeidon.deco.memoize(1000)
    def _text_to_markup(self, text):
        """Return `text` renderer as markup for display."""
        if not text: return ""
        has_comma = text.find(",") > 0
        if has_comma:
            text = text.replace(",", ".")
        text = self._format.format(float(text))
        if has_comma:
            text = text.replace(".", ",")
        return GLib.markup_escape_text(text)



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Cell renderer for multiline text data."""

import aeidon
import difflib
import gaupol
import re

from aeidon.i18n import _
from gi.repository import Gdk
from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gtk

__all__ = ("MultilineCellRenderer", "MultilineDiffCellRenderer")


class CellTextView(Gtk.TextView, Gtk.CellEditable):

    """A :class:`Gtk.TextView` suitable for cell renderer use."""

    __gproperties__ = {
        "editing-canceled": (bool,
                             "Editing canceled",
                             "Editing canceled",
                             False,
                             GObject.ParamFlags.READWRITE),

    }
    __gtype_name__ = "CellTextView"

    def __init__(self, text_buffer=None):
        """Initialize a :class:`CellTextView` instance."""
        GObject.GObject.__init__(self)
        gaupol.util.prepare_text_view(self)
        aeidon.util.connect(self, self, "key-press-event")
        aeidon.util.connect(self, self, "populate-popup")

    def do_editing_done(self, *args):
        """End editing."""
        pass

    def do_remove_widget(self, *args):
        """Remove widget."""
        pass

    def do_start_editing(self, *args):
        """Start editing."""
        # Don't let anyone else handle button-press-events
        # that happen within the text view.
        self.connect_after("button-press-event", lambda *args: True)

    def get_text(self):
        """Return text."""
        text_buffer = self.get_buffer()
        start, end = text_buffer.get_bounds()
        return text_buffer.get_text(start, end, False)

    def _on_key_press_event(self, text_view, event):
        """Toggle italicization if Control+I pressed."""
        if (event.get_state() & Gdk.ModifierType.CONTROL_MASK and
            event.keyval in (Gdk.KEY_i, Gdk.KEY_I)):
            self._on_toggle_italic()
            return True
        return False

    def _on_populate_popup(self, text_view, popup):
        """Add a context menu item to toggle italicization."""
        if not isinstance(popup, Gtk.Menu): return
        item = Gtk.MenuItem(label=_("Italic"))
        item.connect("activate", self._on_toggle_italic)
        child = item.get_child()
        if isinstance(child, Gtk.AccelLabel):
            child.set_accel(Gdk.KEY_i, Gdk.ModifierType.CONTROL_MASK)
        item.show_all()
        popup.append(item)

    def _on_toggle_italic(self, *args):
        """Add or remove italic tags around selection."""
        if not getattr(gaupol, "italic_tag", None): return
        if not getattr(gaupol, "italicize", None): return
        text_buffer = self.get_buffer()
        start, end = text_buffer.get_bounds()
        text = text_buffer.get_text(start, end, False)
        bounds = text_buffer.get_selection_bounds()
        if not bounds:
            # If no selection, replicate cursor position.
            cursor = text_buffer.get_insert()
            cursor = text_buffer.get_iter_at_mark(cursor)
            bounds = (cursor, cursor)
        a, b = [x.get_offset() for x in bounds]
        # Strip possible existing italic tags around selection.
        # If none found, then selection should be italicized.
        text1 = re.sub(gaupol.italic_tag.pattern + r"\Z", "", text[:a])
        text2 = re.sub(gaupol.italic_tag.pattern, "", text[a:b])
        text3 = re.sub(r"\A" + gaupol.italic_tag.pattern, "", text[b:])
        if text1 + text2 + text3 == text:
            text2 = gaupol.italicize(text2)
            match = gaupol.italic_tag.match(text2)
            a += match.end() - match.start()
            b += match.end() - match.start()
        else:
            a = len(text1)
            b = len(text1) + len(text2)
        text_buffer.set_text(text1 + text2 + text3)
        ins = text_buffer.get_iter_at_offset(a)
        bound = text_buffer.get_iter_at_offset(b)
        text_buffer.select_range(ins, bound)

    def set_text(self, text):
        """Set text."""
        self.get_buffer().set_text(text)


class MultilineCellRenderer(Gtk.CellRendererText):

    """
    Cell renderer for multiline text data.

    If :attr:`gaupol.conf.editor.show_lengths_cell` is ``True``, line lengths
    are shown at the end of each line.
    """

    __gtype_name__ = "MultilineCellRenderer"

    def __init__(self):
        """Initialize a :class:`MultilineCellRenderer` instance."""
        GObject.GObject.__init__(self)
        self._in_editor_menu = False
        self._show_lengths = gaupol.conf.editor.show_lengths_cell
        self._text = ""
        gaupol.conf.connect_notify("editor", "show_lengths_cell", self)
        aeidon.util.connect(self, self, "notify::text")

    def do_start_editing(self, event, widget, path, bg_area, cell_area, flags):
        """Initialize and return a :class:`CellTextView` widget."""
        editor = CellTextView()
        gaupol.style.use_font(editor, "custom")
        editor.set_text(self._text)
        editor.set_size_request(cell_area.width, cell_area.height)
        editor.set_left_margin(self.props.xpad)
        editor.set_right_margin(self.props.xpad)
        with aeidon.util.silent(AttributeError):
            # Top and bottom margins available since GTK 3.18.
            editor.set_top_margin(self.props.ypad)
            editor.set_bottom_margin(self.props.ypad)
        editor.gaupol_path = path
        editor.connect("focus-out-event", self._on_editor_focus_out_event)
        editor.connect("key-press-event", self._on_editor_key_press_event)
        editor.connect("populate-popup",  self._on_editor_populate_popup)
        editor.show()
        return editor

    def _on_conf_editor_notify_show_lengths_cell(self, *args):
        """Hide or show line lengths if ``conf`` changed."""
        self._show_lengths = gaupol.conf.editor.show_lengths_cell

    def _on_editor_focus_out_event(self, editor, *args):
        """End editing."""
        if self._in_editor_menu: return
        editor.remove_widget()
        self.emit("editing-canceled")

    def _on_editor_key_press_event(self, editor, event):
        """End editing if ``Enter`` or ``Escape`` pressed."""
        if (event.get_state() &
            (Gdk.ModifierType.SHIFT_MASK |
             Gdk.ModifierType.CONTROL_MASK)): return
        if event.keyval in (Gdk.KEY_Return, Gdk.KEY_KP_Enter):
            editor.editing_done()
            editor.remove_widget()
            self.emit("edited", editor.gaupol_path, editor.get_text())
            return True
        if event.keyval == Gdk.KEY_Escape:
            editor.editing_done()
            editor.remove_widget()
            self.emit("editing-canceled")
            return True

    def _on_editor_populate_popup(self, editor, menu):
        """Disable "focus-out-event" ending editing."""
        self._in_editor_menu = True
        def on_menu_unmap(menu, self):
            self._in_editor_menu = False
        menu.connect("unmap", on_menu_unmap, self)

    def _on_notify_text(self, *args):
        """Set markup by adding line lengths to text."""
        # Since GTK 3.6, the notify::text signal seems to get
        # emitted insanely often even if text hasn't changed at
        # all. Let's try to keep this callback as fast as possible.
        self._text = self.props.text
        self.props.markup = self._text_to_markup(self.props.text,
                                                 self._show_lengths,
                                                 gaupol.conf.editor.length_unit)

    def set_show_lengths(self, show_lengths):
        """Show or hide line lengths, overriding ``conf``."""
        self._show_lengths = show_lengths
        gaupol.conf.disconnect_notify("editor", "show_lengths_cell", self)

    @aeidon.deco.memoize(1000)
    def _text_to_markup(self, text, show_lengths, length_unit):
        """Return `text` rendered as markup for display."""
        # We don't actually use the length_unit argument,
        # but do need it accounted for in memoized values.
        if not text: return ""
        text = GLib.markup_escape_text(text)
        if not show_lengths: return text
        lines = text.split("\n")
        lengths = gaupol.ruler.get_lengths(text)
        return "\n".join(("{} <small>[{:d}]</small>"
                          .format(lines[i], lengths[i])
                          if lines[i] else lines[i]
                          for i in range(len(lines))))


class MultilineDiffCellRenderer(MultilineCellRenderer):

    """Cell renderer for multiline diffed text data."""

    __gtype_name__ = "MultilineDiffCellRenderer"

    ref_text = GObject.Property(type=str, default="")
    ref_type = GObject.Property(type=int, default=-1)

    COLOR_CHANGE = gaupol.conf.general.diff_color_change
    COLOR_DELETE = gaupol.conf.general.diff_color_delete
    COLOR_INSERT = gaupol.conf.general.diff_color_insert

    def __init__(self):
        """Initialize a :class:`MultilineDiffCellRenderer` instance."""
        MultilineCellRenderer.__init__(self)
        self._ref_text = ""

    def _add_diff_markup(self, a, b):
        """Return `a` with markup for parts different from `b`."""
        matcher = difflib.SequenceMatcher(a=a, b=b)
        ops = matcher.get_opcodes()
        return "".join(self._highlight(a[i1:i2], tag)
                       for tag, i1, i2, j1, j2 in ops)

    def _get_diff_color(self, tag):
        """Return color to use to highlight changes of type `tag`."""
        if tag == "insert" and self.props.ref_type > 0: return self.COLOR_INSERT
        if tag == "insert" and self.props.ref_type < 0: return self.COLOR_DELETE
        if tag == "delete" and self.props.ref_type > 0: return self.COLOR_DELETE
        if tag == "delete" and self.props.ref_type < 0: return self.COLOR_INSERT
        return self.COLOR_CHANGE

    def _highlight(self, text, tag):
        """Return `text` as markup highlighting changes of type `tag`."""
        if tag == "equal": return text
        color = self._get_diff_color(tag)
        return '<span background="{}">{}</span>'.format(color, text)

    def _on_notify_text(self, *args):
        """Set markup by adding line lengths to text."""
        # Since GTK 3.6, the notify::text signal seems to get
        # emitted insanely often even if text hasn't changed at
        # all. Let's try to keep this callback as fast as possible.
        self._text = self.props.text
        self._ref_text = self.props.ref_text
        self.props.markup = self._text_to_markup(self.props.text,
                                                 self.props.ref_text,
                                                 self._show_lengths,
                                                 gaupol.conf.editor.length_unit)

    @aeidon.deco.memoize(1000)
    def _text_to_markup(self, text, ref_text, show_lengths, length_unit):
        """Return `text` rendered as markup for display."""
        # We don't actually use the length_unit argument,
        # but do need it accounted for in memoized values.
        if not text: return ""
        text = GLib.markup_escape_text(text)
        ref_text = GLib.markup_escape_text(ref_text)
        with aeidon.util.silent(Exception, tb=True):
            text = self._add_diff_markup(text, ref_text)
        if not show_lengths: return text
        lines = text.split("\n")
        lengths = gaupol.ruler.get_lengths(text)
        return "\n".join(("{} <small>[{:d}]</small>"
                          .format(lines[i], lengths[i])
                          if lines[i] else lines[i]
                          for i in range(len(lines))))



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Dialog for lengthening or shortening durations."""

import gaupol

from aeidon.i18n   import _, n_
from gi.repository import Gtk

__all__ = ("DurationAdjustDialog",)


class DurationAdjustDialog(gaupol.BuilderDialog):

    """Dialog for lengthening or shortening durations."""

    _widgets = [
        "all_radio",
        "current_radio",
        "gap_check",
        "gap_spin",
        "lengthen_check",
        "max_check",
        "max_spin",
        "min_check",
        "min_spin",
        "selected_radio",
        "shorten_check",
        "speed_spin",
    ]

    def __init__(self, parent, application):
        """Initialize a :class:`DurationAdjustDialog` instance."""
        gaupol.BuilderDialog.__init__(self, "duration-adjust-dialog.ui")
        self.application = application
        self._init_dialog(parent)
        self._init_values()
        self._init_sensitivities()

    def _adjust_durations(self):
        """Adjust durations of subtitles."""
        conf = gaupol.conf.duration_adjust
        target = self._get_target()
        for page in self.application.get_target_pages(target):
            self.application.set_current_page(page)
            rows = page.project.adjust_durations(
                indices=self.application.get_target_rows(target),
                speed=conf.speed,
                lengthen=conf.lengthen,
                shorten=conf.shorten,
                maximum=(conf.maximum if conf.use_maximum else None),
                minimum=(conf.minimum if conf.use_minimum else None),
                gap=conf.gap if conf.use_gap else None)
            self.application.flash_message(n_(
                "Adjusted duration of {:d} subtitle",
                "Adjusted durations of {:d} subtitles",
                len(rows)).format(len(rows)))

    def _get_target(self):
        """Return the selected target."""
        if self._selected_radio.get_active():
            return gaupol.targets.SELECTED
        if self._current_radio.get_active():
            return gaupol.targets.CURRENT
        if self._all_radio.get_active():
            return gaupol.targets.ALL
        raise ValueError("Invalid target radio state")

    def _init_dialog(self, parent):
        """Initialize the dialog."""
        self.add_button(_("_Cancel"), Gtk.ResponseType.CANCEL)
        self.add_button(_("_Adjust"), Gtk.ResponseType.OK)
        self.set_default_response(Gtk.ResponseType.OK)
        self.set_transient_for(parent)
        self.set_modal(True)

    def _init_sensitivities(self):
        """Initialize sensitivities of widgets."""
        self._all_radio.emit("toggled")
        self._current_radio.emit("toggled")
        self._gap_check.emit("toggled")
        self._gap_spin.emit("value-changed")
        self._lengthen_check.emit("toggled")
        self._max_check.emit("toggled")
        self._max_spin.emit("value-changed")
        self._min_check.emit("toggled")
        self._min_spin.emit("value-changed")
        self._selected_radio.emit("toggled")
        self._shorten_check.emit("toggled")
        self._speed_spin.emit("value-changed")

    def _init_values(self):
        """Intialize default values for widgets."""
        conf = gaupol.conf.duration_adjust
        self._gap_check.set_active(conf.use_gap)
        self._gap_spin.set_value(conf.gap)
        self._lengthen_check.set_active(conf.lengthen)
        self._max_check.set_active(conf.use_maximum)
        self._max_spin.set_value(conf.maximum)
        self._min_check.set_active(conf.use_minimum)
        self._min_spin.set_value(conf.minimum)
        self._shorten_check.set_active(conf.shorten)
        self._speed_spin.set_value(conf.speed)
        self._selected_radio.set_active(conf.target == gaupol.targets.SELECTED)
        self._current_radio.set_active(conf.target == gaupol.targets.CURRENT)
        self._all_radio.set_active(conf.target == gaupol.targets.ALL)
        page = self.application.get_current_page()
        rows = page.view.get_selected_rows()
        if not rows and conf.target == gaupol.targets.SELECTED:
            self._current_radio.set_active(True)
        self._selected_radio.set_sensitive(bool(rows))

    def _on_gap_check_toggled(self, check_button):
        """Set sensitivity of the gap spin button."""
        self._gap_spin.set_sensitive(check_button.get_active())

    def _on_lengthen_check_toggled(self, *args):
        """Set sensitivity of the speed spin button."""
        lengthen = self._lengthen_check.get_active()
        shorten = self._shorten_check.get_active()
        self._speed_spin.set_sensitive(lengthen or shorten)

    def _on_max_check_toggled(self, check_button):
        """Set sensitivity of the maximum spin button."""
        self._max_spin.set_sensitive(check_button.get_active())

    def _on_min_check_toggled(self, check_button):
        """Set sensitivity of the minimum spin button."""
        self._min_spin.set_sensitive(check_button.get_active())

    def _on_response(self, dialog, response):
        """Save settings and adjust durations."""
        conf = gaupol.conf.duration_adjust
        conf.gap = self._gap_spin.get_value()
        conf.lengthen = self._lengthen_check.get_active()
        conf.maximum = self._max_spin.get_value()
        conf.minimum = self._min_spin.get_value()
        conf.speed = self._speed_spin.get_value()
        conf.shorten = self._shorten_check.get_active()
        conf.target = self._get_target()
        conf.use_gap = self._gap_check.get_active()
        conf.use_maximum = self._max_check.get_active()
        conf.use_minimum = self._min_check.get_active()
        if response == Gtk.ResponseType.OK:
            self._adjust_durations()

    def _on_shorten_check_toggled(self, *args):
        """Set sensitivity of the speed spin button."""
        lengthen = self._lengthen_check.get_active()
        shorten = self._shorten_check.get_active()
        self._speed_spin.set_sensitive(lengthen or shorten)
